#include <iostream>
#include <cmath>
#include <fstream>
#include <string>
#include <vector>

using namespace std;

ofstream debuglog;
ofstream colorchecklog;

// ios binary means DON'T try and fix unix line endings / etc.

// technically all bytes
float find_magnitude(float ir, float ig, float ib)
{

    float tTemp1 = ir*ir + ig*ig + ib*ib;
    debuglog << tTemp1 << endl;
    float tTemp2 = sqrt(tTemp1);
    debuglog << tTemp2 << endl;
    return tTemp2;
}

struct RGB
{
    unsigned char R;
    unsigned char G;
    unsigned char B;
};

struct Pal
{
    RGB Colors[256];
};

struct PLAYPAL
{
    Pal Palettes[14];
};

// add file open error
PLAYPAL ReadPaletteFile(string iFileName)
{
    PLAYPAL tPal;
    ifstream infile;
    infile.open(iFileName.c_str(), ios::binary);
    if(!infile)
    {
        debuglog << "Error: Could not read " << iFileName << endl;
    }
    infile.read((char*)&tPal, sizeof(PLAYPAL));
    debuglog << "sizeof(PLAYPAL) is " << sizeof(PLAYPAL) << endl;

    return tPal;
}

bool WritePaletteFile(PLAYPAL iPal, string iFileName)
{
    ofstream tempfile;
    tempfile.open(iFileName.c_str(), ios::binary);
    if(!tempfile)
    {
        return false;
    }

    tempfile.write((char*)&iPal,sizeof(PLAYPAL));
    return true;
}



float min3(float a, float b, float c)
{
    float t1 = min(a,b);
    float t2 = min(b,c);
    return min(t1,t2);
}

float max3(float a, float b, float c)
{
    float t1 = max(a,b);
    float t2 = max(b,c);
    return max(t1,t2);
}

struct HSV
{
    float H;
    float S;
    float V;
};

struct HSV_Pal
{
    HSV Colors[256];
};

// potential member function of a COLOR class? Convert to HSV?

// code based on http://www.cs.rit.edu/~ncs/color/t_convert.html
HSV RGBtoHSV(RGB iRGB)
{
	float tmin, tmax, delta;
	HSV tHSV;

	tmin = min3( iRGB.R, iRGB.G, iRGB.B );
	tmax = max3( iRGB.R, iRGB.G, iRGB.B );
	tHSV.V = tmax;
	// value is the maximum of r,g, and b, it is brightness

	delta = tmax - tmin;

	if( tmax != 0 )
		tHSV.S = delta / tmax;		// s

	else
    {
		// r = g = b = 0		// s = 0, v is undefined
		tHSV.S = 0;
		tHSV.H = -1;
		return tHSV;
	}

	if( iRGB.R == tmax )
		tHSV.H = ( iRGB.G - iRGB.B ) / delta;		// between yellow & magenta
	else if( iRGB.G == tmax )
		tHSV.H = 2 + ( iRGB.B - iRGB.R ) / delta;	// between cyan & yellow
	else
		tHSV.H = 4 + ( iRGB.R - iRGB.G ) / delta;	// between magenta & cyan

	tHSV.H *= 60;				// degrees
	if( tHSV.H < 0 )
		tHSV.H += 360;

    return tHSV;

}

    typedef vector<RGB> SearchResults;

void PrintHSV(HSV iHSV)
{
    debuglog << "H is " << iHSV.H << endl;
    debuglog << "S is " << iHSV.S << endl;
    debuglog << "V is " << iHSV.V << endl;
    return;
}

void LogHSV(HSV iHSV)
{
    debuglog << "H is " << iHSV.H << endl;
    debuglog  << "S is " << iHSV.S << endl;
    debuglog  << "V is " << iHSV.V << endl;
    return;
}

void PrintRGB(RGB iRGB)
{
    debuglog << "R is " << (int) iRGB.R << endl;
    debuglog << "G is " << (int) iRGB.G << endl;
    debuglog << "B is " << (int) iRGB.B << endl;
    return;
}

struct ColorMapRow
{
    int Indexes[34];
};

struct RGBColorMapRow
{
    RGB Colors[34];
};

struct Color_Compare
{
    float Delta_H;
    float Delta_V;
    float Delta_S;
    int Index;
};

void PrintColorMapRow(ColorMapRow iColorMapRow)
{
    for(int i = 0; i < 34; i++)
    {
        debuglog << "Colormaprow[" << i << "] == " << iColorMapRow.Indexes[i] << endl;
    }

    return;
}

void LogColorMapRow(ColorMapRow iColorMapRow)
{
    debuglog << "----LogColorMapRow----" << endl;
    for(int i = 0; i < 34; i++)
    {
        debuglog << "Colormaprow[" << i << "] == " << iColorMapRow.Indexes[i] << endl;
    }

    return;
}

void LogRGBColorMapRow(RGBColorMapRow iColorMapRow)
{
    debuglog << "----LogRGBColorMapRow---" << endl;
    for(int i = 0; i < 34; i++)
    {
        debuglog << "i = " << i << endl;
        PrintRGB(iColorMapRow.Colors[i]);
    }

    return;
}

// step can be adjusted if you want a more colorful palette in the low range
// calculate the step value to get from iColor to black in iNumberOfRows rows.
RGB CalcStep(RGB iColor, int iNumberOfRows)
{
    RGB tStep;

    // safely handling divide by zero errors.
    if(iNumberOfRows == 0)
    {
        tStep.R = 254;
        tStep.G = 254;
        tStep.B = 254;

        return tStep;
    }

    tStep.R = iColor.R/iNumberOfRows;
    tStep.G = iColor.G/iNumberOfRows;
    tStep.B = iColor.B/iNumberOfRows;

    return tStep;
}


// NOTE, this fills in dummy (white) values for 33 and 34
RGBColorMapRow MakeRGBGradient(RGB iColor1)
{
    RGBColorMapRow tColorMapRow;
    RGB tStep;

    tStep = CalcStep(iColor1, 32);

    for(int i = 0; i < 32 ; i++)
    {
        tColorMapRow.Colors[i].R = iColor1.R - tStep.R*i;
        tColorMapRow.Colors[i].G = iColor1.G - tStep.G*i;
        tColorMapRow.Colors[i].B = iColor1.B - tStep.B*i;
    }

    tColorMapRow.Colors[32].R = 255;
    tColorMapRow.Colors[32].G = 255;
    tColorMapRow.Colors[32].B = 255;

    tColorMapRow.Colors[33].R = 255;
    tColorMapRow.Colors[33].G = 255;
    tColorMapRow.Colors[33].B = 255;


    return tColorMapRow;
}


int FindNearestIndex(RGB iRGB, Pal iPalette)
{
    colorchecklog << "----------START FindNearestIndex---------" << endl;
    colorchecklog << "iRGB.R = " << (int) iRGB.R << endl;
    colorchecklog << "iRGB.G = " << (int) iRGB.G << endl;
    colorchecklog << "iRGB.B = " << (int) iRGB.B << endl;


	int tMin_Diff = 999999;
	int tBestMatchIndex = 0;
	int tDiff;
	RGB tRGB_Delta;

	for (int i = 0; i < 256; i++)
	{
		tRGB_Delta.R = iRGB.R - iPalette.Colors[i].R;
		tRGB_Delta.G = iRGB.G - iPalette.Colors[i].G;
		tRGB_Delta.B = iRGB.B - iPalette.Colors[i].B;
		tDiff = (tRGB_Delta.R*tRGB_Delta.R)+(tRGB_Delta.G*tRGB_Delta.G)+(tRGB_Delta.B*tRGB_Delta.B);

		// If the current color in PLAYPAL is an exact match with iRGB,
		// return the index of that current color.
		if (tDiff == 0)
		{
			tBestMatchIndex = i;
			colorchecklog << "Exact match found with index " << tBestMatchIndex << endl;
		    colorchecklog << "With RGB {" << (int)iPalette.Colors[i].R << ',' << (int)iPalette.Colors[i].G << ',' << (int)iPalette.Colors[i].B << '}' << endl << endl;

			return tBestMatchIndex;
		}

		else if (tDiff < tMin_Diff)
		{
		    colorchecklog << "Best Match updated " << endl;
		    colorchecklog << "Was tBestMatchIndex = " << (int)tBestMatchIndex << " with tMin_Diff of " << (int)tMin_Diff << endl;
		    colorchecklog << "Updated to " << (int)i << " with tDiff of " << (int)tDiff << endl;
		    colorchecklog << "With RGB {" << (int)iPalette.Colors[i].R << ',' << (int)iPalette.Colors[i].G << ',' << (int)iPalette.Colors[i].B << '}' << endl << endl;
			tMin_Diff = tDiff;
			tBestMatchIndex = i;
		}

		colorchecklog << "index " << i << endl;
        colorchecklog << "With RGB {" << (int)iPalette.Colors[i].R << ',' << (int)iPalette.Colors[i].G << ',' << (int)iPalette.Colors[i].B << '}' << endl;
        colorchecklog << "diff = " << (int) tDiff << endl << endl;
	}
	// we have not found any perfect matches to iRGB, and we have
	// went through the entire palette. Return the best match we've found.
	return tBestMatchIndex;
}

// a "row" in this case is the full set of 32 indexes for a given color.
ColorMapRow ConvertRGBRowToIndexRow(RGBColorMapRow iRGBColorMap, Pal iPalette)
{
    // set to find the best match in R, G, and B.
    // if you weigh one dimension over another you will get different results.
    ColorMapRow tColorMapRow;

    for(int tCMRow = 0; tCMRow < 34; tCMRow++)
    {
		int tSearchResult = 255;
		tSearchResult = FindNearestIndex(iRGBColorMap.Colors[tCMRow],iPalette);

        tColorMapRow.Indexes[tCMRow] = tSearchResult;
    }

    return tColorMapRow;
}


ColorMapRow MakeColorMapRow(Pal iPalette, RGB iColor)
{
// 1. make 32 pix gradient from iColor to black
    RGBColorMapRow tRGBColorMapRow = MakeRGBGradient(iColor);
    LogRGBColorMapRow(tRGBColorMapRow);

// 2. For each pixel in the gradient, find the nearest match to iPalette
    ColorMapRow tColorMapRow = ConvertRGBRowToIndexRow(tRGBColorMapRow,iPalette);
    LogColorMapRow(tColorMapRow);

    return tColorMapRow;

}

// there are 34 rows of colormaps. 32 for brightness and 2 for special cases
// the COLORMAP file is 256(W) x 32(H)
class COLORMAP_File
{
    public:
    unsigned char indexes[34][256]; //34 rows of 256 colors.

    // note that iRowIndex means the color to map these indexes to
    // runs from 0 to 255
    void ColorMapRowToColorMap(int iRowColorIndex, ColorMapRow iColorMapRow)
    {
        for(int t = 0; t < 34; t++)
        {
            indexes[t][iRowColorIndex] = iColorMapRow.Indexes[t];
        }
    }
};



bool WriteCOLORMAPFile(COLORMAP_File iCM, string iFileName)
{
    ofstream tempfile;
    tempfile.open(iFileName.c_str(), ios::binary);
    if(!tempfile)
    {
        return false;
    }

    tempfile.write((char*)&iCM,sizeof(COLORMAP_File));
    return true;
}

int main()
{
    // general debugging
    debuglog.open("log.txt");

    // monitors the conversion from RGB gradients to indexed color gradients.
    colorchecklog.open("colorcheck.txt");

    PLAYPAL tPal;
    tPal = ReadPaletteFile("C:\\PAL\\PLAYPAL.pal");
    WritePaletteFile(tPal, "output.pal");

    COLORMAP_File tColormap_file;

    /*

    ColorMapRow tColorMapRow;
    tColorMapRow = MakeColorMapRow(tPal.Palettes[0],tPal.Palettes[0].Colors[16]);
    // temporarily write the same row data to all rows.
    for(int i = 0; i < 256; i++)
    {
        tColormap_file.ColorMapRowToColorMap(i,tColorMapRow);
    }
    */

    // the real deal

    for(int i = 0; i < 256; i++)
    {
        cout << "Row " << i << " of 255" << endl;
        ColorMapRow tColorMapRow;
        tColorMapRow = MakeColorMapRow(tPal.Palettes[0],tPal.Palettes[0].Colors[i]);
        tColormap_file.ColorMapRowToColorMap(i,tColorMapRow);
    }



    WriteCOLORMAPFile(tColormap_file,"Colormaptest.cm");





    return 0;
}
